#include "local.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include "son.h"

#ifdef ALLOC_H
#include <alloc.h>
#endif

#ifdef MALLOC_H
#include <malloc.h>
#endif

#ifdef NO_MALLOC_H
/* 
** malloc.h is not defined on the Apollo machines 
*/
	extern char *malloc(unsigned);
	extern int free(char *);
	extern char *remalloc(char *,unsigned);
	extern char *calloc(unsigned,unsigned);
	extern int cfree(char *);
	extern char *memalign(unsigned,unsigned);
	extern char *valloc(unsigned);

#endif /* NO_MALLOC_H */


#define NEW_CLEAR 0		
#define NEW_TRACE 0

/******************************************************************

NEW TRACE LEVEL DEFINITION:
LEVEL 1: NEW_DBG

	Determines if memory allocation is good.
	Make sure that the next location after the allocated area is 
	not change.  Keeps trace of:
		1. size of allocated memory
		2. a uniq number which is incremented every time an allocation
		   is done, including new, rnew and cnew.  May be used to 
		   trace when an allocation is done.  To do so break in
		   "memory_nu" when alloc_nu == the allocation number.
		3. File and line where allocation was done.
		4. Address of memory being allocated.
		5. Total amount of memory being allocated.

Level 2: NEW_CLEAR

	Make sure that uninitialize memory is not used.  

	new():  For a new() set allocation memory allocated to FILL.
	rnew(): Create new memory using a new().  Set memory to FILL, then copy 
		over the previous memory.  Next set original memory to FILL,
		then free the allocated memory.
	free(): Set all memory to FILL then free it.

Level 3: NEW_TRACE

	Prints out all allocations.
	where n = the uniq. allocation number.

	SET IN sonlib.h
NOTE:
	1. Production code should have all debugging turned off.
	2. To use NEW_CLEAR or NEW_TRACE must have NEW_DBG turn on.

BASIC ROUTINES:
	new() - allocates memory
	cnew() - allocates clear memory
	rnew() - reallocates memory
	sFree() - frees allocated memory
	init_new() - initialize new allocation. - See DEBUG ROUTINES

DEBUG ROUTINES:
	print_all_new 
		prints out all allocations

	print_new(n) 
		prints out a specific allocation

	set_new_trace_level(level) 
		sets trace level (1, 2, or 3)

	set_new_trace_size(min) 
		using trace level 3, only prints out memory allocations 
		(or free) if amount being allocated is larger than min.

	set_new_break_function(i,func) 
		calls function func when doing the ith allocation.  Use in case
		we trace allocation is being done that is causing
		a problem.  We want to do a break point on this allocation..
		Set the break point in the function "func".

	new_print_memory_alloc()
		print out total amount of memory allocation already done.

	init_new()
		Set enviornment variable DNEW to
		a) number - trace level number
		b) number size - trace level number (must be 3 or above,
			print out all allocations larger than size - in 
			Kilobytes.  If value is 0 then or trace level is
			2 or below, then size is ignored.
		c) number size msg - msg is printed out when init_new is called
		   to system error.

******************************************************************/

#undef new
#undef rnew
#undef cnew
#undef sFree
#undef stop_if_nomem

#ifdef DOS
#define ADDR	long int
#else
#define ADDR	char *
#endif

static int total = 0;
static int stop_on_nomem = 1;
static int make_core = 0;
static int trace = 0;

#define END_CHAR	'X'
#define FILL		'Y'

#if defined(IBM6000) || defined(HP700) || defined(ALPHA)
void *bsearch();
#else
char *bsearch();
#endif
void qsort();
static int alloc_nu = 1;
static char **ptr_tbl=NULL;
static int *ptr_size=NULL;
static int *ptr_nu=NULL;
static char **ptr_file=NULL;
static int *ptr_linenu=NULL;
static int nel=0;
static char *filename = "unknown";
static int line_number = 0;
static int min_print_size = 0;
static void (*trace_func)() = NULL;
static int new_alloc_nu = -1;

static int cmp_ptr(
#if NeedFunctionPrototypes
	char **x,
	char **y
#endif
);

static void ptr_tbl_add(
#if NeedFunctionPrototypes
	char *x,
	int size
#endif
);

static void
ptr_sort(
#if NeedFunctionPrototypes
	void
#endif
);

static void
print_new(
#if NeedFunctionPrototypes
	int x
#endif
);

static void 
memory_nu(
#if NeedFunctionPrototypes
	int alloc_nu
#endif
);

static void
new_error(
#if NeedFunctionPrototypes
	char *msg
#endif
);

/*************************************************************/
/* Set trace level depending on DNEW environment varaible     */
/*************************************************************/
void 
init_new()
{
	extern char *getenv();
	char *dnew;
	char msg[60];
	int trace_level, size;

	dnew = getenv("DNEW");
	if ( dnew == NULL ) return;

	sscanf(dnew,"%d %d %s",&trace_level,&size,msg);
	if ( trace_level <=2) size = 0;
	if ( trace_level <0 || trace_level >3 ) {
		fprintf(stderr,"new: DNEW = %s\n",dnew);
		fprintf(stderr,"Trace level must be 1, 2 or 3\n");
		exit_program(1);
	}
	if ( trace_level > 0 ) {
		set_new_trace_level(trace_level);
	}

	if ( size < 0 ) {
		fprintf(stderr,"new: DNEW = %s\n",dnew);
		fprintf(stderr,"Size must be an integer greater than 1\n");
		fprintf(stderr,"size =  %d\n",size);
		exit_program(1);
	}
	if ( size > 0 ) {
		set_new_trace_size(size*1024);
	}
	if ( strcmp(msg,"") != 0 ) {
		fprintf(stderr,"DEBUG: %s, Trace level %d, Trace size %d\n",
			msg, trace_level,size);
	}
}

/*************************************************************/
/* Set the trace level (0,1,2,3) - must be set before any    */
/* new.c functions are called.  By default trace is set off  */
/*************************************************************/
void 
set_new_trace_level(level)
int level;
{
	trace = level;
}

void
/*************************************************************/
/* Set the size of the malloc to be printed out.  Only print */
/* out mallocs, free which are larger than the specified one */
/* Need to have set trace level to 3 or above to use this    */
/* feature						     */
/* new.c functions are called.  By default trace is set off  */
/*************************************************************/
set_new_trace_size(min)
int min;
{
	if ( trace <=2 ) {
		announce_bug("set_new_trace_size: Must have set_new_trace_level set to 3 or above");
	}
	min_print_size = min;
}

void
/*************************************************************/
/* Set the size of the malloc to be printed out.  Only print */
/* out mallocs, free which are larger than the specified one */
/* Need to have set trace level to 3 or above to use this    */
/* feature						     */
/* new.c functions are called.  By default trace is set off  */
/*************************************************************/
set_new_break_function(alloc_nu,func)
int alloc_nu;
void (*func)();
{
	if ( trace <= 1 ) {
		announce_bug("set_new_break_function: Must have set_new_trace_level set to 2 or above");
	}
	new_alloc_nu = alloc_nu;
	trace_func = func;
}

char *new(size) unsigned size;
/*************************************************************/
/* Allocate memory with malloc, if error exit program.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11mar88 */ 
{ 
	char *A;

	if ( trace >= 1 ) {
		if ( size <= 0 ) {
			fprintf(stderr,"Invalid memory allocation for new() = %d\n",size);
			exit_program(1);
		}
		++size;
	}

	if ( (A=malloc(size)) == NULL ) {  
		if ( stop_on_nomem ) { 
			new_error( "Out of memory (malloc).\n");
		} else { 

			if (trace>=3) printf("new: %s:%d size=%d, OUT OF MEMORY\n",
					     filename,line_number,size);
			return (NULL);
		}
	}


	if ( trace >= 1 ) {
		new_check();
		A[--size] = END_CHAR;
		total += size;
		if (trace>=3) {
			if ( size > min_print_size ) {
				printf("new: %s:%d size=%d, addr=%lx\n",
					     filename,line_number,size,A);
			}
		}
		if (trace>=2) memset(A,FILL,size);

		ptr_tbl_add(A,size);
	}

	return(A);
} /*new*/

char *rnew(ptr,size) char *ptr;  unsigned size;
/*************************************************************/
/* Allocate memory with malloc, if error exit program.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11mar88 */ 
{ 
	char *A;

	int i;
	char *tbl;
	char **loc;
	int next;
	int num;

	if (trace>=1) {
		if ( size <= 0 ) {
			fprintf(stderr,"Invalid memory allocation for rnew() = %d\n",size);
			exit_program(1);
		}
		++size;
		new_check();
	}

	if ( trace>=2) {
		A=malloc(size);
	} else {
		A=(char *)realloc(ptr,size);
	}

	if ( A == NULL ) { 
		if ( stop_on_nomem ) {
			new_error( "Out of memory (realloc).\n");
		} else { 
			if (trace>=3) {
				printf("rnew: %s:%d size=%d, OUT OF MEMORY\n",
						     filename,line_number,size);
			}
			return (NULL);
		}
	}

	if ( trace >= 1) {
		if (trace>=3) {
			if ( size > min_print_size ) {
				printf("rnew: %s:%d size=%d oldaddr=%lx addr=%lx\n",
				      filename,line_number,size,ptr,A);
			}
		}

		loc = (char **)bsearch((char *)(&ptr),(char *)ptr_tbl,
		nel,sizeof(ptr),cmp_ptr);
		next = (int)(loc - ptr_tbl);
		total += (size-ptr_size[next]-1);

		if (loc==NULL) {
			printf("rnew: size=%d oldaddr=%lx addr=%lx\n",
				size,ptr,A);
			print_new(-1);
			announce_bug("rnew");
		}
		*loc = A;

		if ( trace>=2) {
			memset(A,FILL,size);
			if ( size < ptr_size[next] ) {
				num = size;
			} else {
				num = ptr_size[next];
			}
			memcpy(A,ptr,num);
			memset(ptr,FILL,ptr_size[next]);
			free(ptr);
		}
		A[--size] = END_CHAR;

		tbl = ptr_tbl[next];

		free(ptr_file[next]);
		for ( i = next; i < nel-1;  ++i ) {
			ptr_tbl[i] = ptr_tbl[i+1];
			ptr_size[i] = ptr_size[i+1];
			ptr_nu[i] = ptr_nu[i+1];
			ptr_file[i] = ptr_file[i+1];
			ptr_linenu[i] = ptr_linenu[i+1];
		}

		memory_nu(alloc_nu);
		ptr_tbl[nel-1] = tbl;
		ptr_size[nel-1] = size;
		ptr_file[nel-1] = malloc((strlen(filename)+1)*sizeof(char));
		strcpy( ptr_file[nel-1],filename);
		ptr_linenu[nel-1] = line_number;
		ptr_nu[nel-1] = alloc_nu++;

		ptr_sort();

		new_check();

	}  /* End trace */

	return(A);

} /*rnew*/

char *cnew(n,size) unsigned n,size;
/*************************************************************/
/* Allocate memory with malloc, if error exit program.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11mar88 */ 
{ 
	char *A;

	if ( trace>=1) {
		if ( size <= 0 ) {
			fprintf(stderr,"Invalid memory allocation for new() = %d\n",size);
			exit_program(1);
		}
		++size;
	}

	if ( (A=(char *)calloc(n,size)) == NULL ) { 
		if ( stop_on_nomem )  {
			new_error("Out of memory (calloc).\n");
		} else { 
			if (trace>=3) printf("cnew: %s:%d size=%d, OUT OF MEMORY\n",
			filename,line_number,size*n);
			return (NULL);
		}
	}

	if ( trace>=1) {
		--size;
		new_check();

		if (trace>=3) {
			if ( size > min_print_size ) {
				printf("cnew: %s:%d size=%d  addr=%lx\n",
					filename,line_number,size*n,A);
			}
		}

		A[n*size] = END_CHAR;
		total += (n*size);

		/* Bug checking tbl */
		ptr_tbl_add(A,n*size);
	}

	return(A);

} /*cnew*/

void sFree(ptr) char *ptr;
/*************************************************************/
/* Release the memory pointed to by ptr.                     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16may89 */ 
{
	char **loc,**loc_stop;
	int *iloc,*iloc_stop;
	char **cloc,**cloc_stop;
	int next;

	if ( trace>=1) {

		new_check();

		loc = (char **)bsearch((char *)(&ptr),(char *)ptr_tbl,
			       nel,sizeof(ptr),cmp_ptr);
		if (loc==NULL) {
			printf("free: %s:%d addr=%lx\n",filename,line_number,ptr);
			announce_bug("sFree");
		}
		next = (int)(loc - ptr_tbl);
		total -= ptr_size[next];
		if (trace>=3) {
			if ( ptr_size[next]> min_print_size) {
				printf("free: %s:%d size=%d addr=%lx (%s:%d)\n",
					filename,line_number,ptr_size[next],
					ptr, ptr_file[next], ptr_linenu[next]);
			}
		}

		if ( trace>=2) {
			/* Clear out the memory */
			memset(ptr,FILL,ptr_size[next]);
		}

		for (loc_stop=ptr_tbl+nel-1; loc<loc_stop; loc++) {
			*loc = *(loc+1);
		}

		iloc = ptr_size + next;
		for (iloc_stop=ptr_size+nel-1; iloc<iloc_stop; iloc++) {
			*iloc = *(iloc+1);
		}

		cloc = ptr_file + next;
		for (cloc_stop=ptr_file+nel-1; cloc<cloc_stop; cloc++) {
			*cloc = *(cloc+1);
		}

		iloc = ptr_linenu + next;
		for (iloc_stop=ptr_linenu+nel-1; iloc<iloc_stop; iloc++) {
			*iloc = *(iloc+1);
		}
		iloc = ptr_nu + next;
		for (iloc_stop=ptr_nu+nel-1; iloc<iloc_stop; iloc++) {
			*iloc = *(iloc+1);
		}
		nel--;

	} /* End trace */

	free(ptr);
  return;
} /*sFree*/

void
stop_if_nomem(value)
int value;
/* krs25may91 */ 
/*************************************************************/
/* Determines if new(), rnew() and cnew() should terminate   */
/* program if no memory is available.  If value is TRUE then */
/* terminate program, else return NULL			     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
{
	stop_on_nomem = value;

	if (trace>=3) printf("stop_if_nomem: %s:%d %s\n",
			filename,line_number, (value)? "On" : "Off" );
}

void
new_core(value) 
int value;
/* krs25may91 */ 
/*************************************************************/
/* If value is TRUE then produce a core file when out of     */
/* memory occurs.					     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
{
	make_core = value;
}

void
core_program()
/* krs25may91 */ 
/*************************************************************/
/* Produce a core file					     */
/* memory occurs.					     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
{
	int a,b;
	a=1;

	/* Make sure we are not stopped by the optimizer compiler */
	for (b = -1; b< 1; ++b ) {
		if (b > 0 ) printf("%d\n",a/b);
	}
}

static int cmp_ptr(x,y)  char **x,**y;
/*************************************************************/
/* Compare two character pointers.                           */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16may89 */ 
{
  if (*x == *y) return(0);
  if ( (ADDR)*x > (ADDR)*y ) {
	return 1;
  } else {
  	return -1;
  }
} /*cmp_ptr*/

static void ptr_tbl_add(x,size)  char *x; int size;
/*************************************************************/
/* Add a character pointer to the table.                     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16may89 */ 
{ static int max=0;

  if (nel>=max)
  { max += 256;
    ptr_tbl = (ptr_tbl==NULL) ? (char **)malloc(max*sizeof(*ptr_tbl))
			      : (char **)realloc(ptr_tbl,max*sizeof(*ptr_tbl));
    ptr_size = (ptr_size==NULL) ? (int *)malloc(max*sizeof(int))
			      : (int *)realloc(ptr_size,max*sizeof(int));
    ptr_file = (ptr_file==NULL) ? (char **)malloc(max*sizeof(char *))
			      : (char **)realloc(ptr_file,max*sizeof(char *));
    ptr_linenu = (ptr_linenu==NULL) ? (int *)malloc(max*sizeof(int))
			      : (int *)realloc(ptr_linenu,max*sizeof(int));
    ptr_nu = (ptr_nu==NULL) ? (int *)malloc(max*sizeof(int))
			      : (int *)realloc(ptr_nu,max*sizeof(int));
  }

  memory_nu(alloc_nu);
  *(ptr_tbl + nel) = x;
  *(ptr_size + nel) = size;
  *(ptr_file + nel) = malloc((strlen(filename)+1)*sizeof(char));
  strcpy( *(ptr_file+nel),filename);
  *(ptr_linenu + nel) = line_number;
  *(ptr_nu + nel) = alloc_nu++;
  nel++;
  ptr_sort();

} /*ptr_tbl_add*/

static void
ptr_sort()
{
	int i,next;
	char *tbl;
	int size;
	int nu;
	int linenu;
	char *file;

	next = nel - 1;
	tbl = ptr_tbl[next];
	size = ptr_size[next];
	file = ptr_file[next];
	linenu = ptr_linenu[next];
	nu = ptr_nu[next];

	for ( i=0; i < nel-1; ++i ) {
		if ( (ADDR)ptr_tbl[i] >  (ADDR)tbl ) {
			next =i;
			break;
		}
	}
	for ( i = nel-1; i > next; --i ) {
		ptr_tbl[i] = ptr_tbl[i-1];
		ptr_size[i] = ptr_size[i-1];
		ptr_file[i] = ptr_file[i-1];
		ptr_linenu[i] = ptr_linenu[i-1];
		ptr_nu[i] = ptr_nu[i-1];
	}
	ptr_tbl[next] = tbl;
	ptr_size[next] = size;
	ptr_file[next] = file;
	ptr_linenu[next] = linenu;
	ptr_nu[next] = nu;

	if (trace>3) print_new(-1);
}

static void
print_new(x)
int x;
{
	int found;
	int i;

	printf("Index  File          line  alloc_nu size   address\n");
	found = 0;
	for (i=0; i < nel; ++i) {
		if ( x == -1 || ptr_nu[i] == x ) {
			printf("%4d %+15s:%6d %5d  %5d   %lx\n",
				i,ptr_file[i],ptr_linenu[i],
				ptr_nu[i],ptr_size[i],ptr_tbl[i] );
		found = 1;
		}
	}
	if ( x == -1 ) {
		printf("\n");
		printf("Total: %d\n",total);
	} else {
		if ( ! found ) {
			printf("Not found\n");
		}
	}
}

void
new_check()
{
	int i;
	int error;
	error = 0;
	for (i=0; i < nel; ++i) {
		if ( *(ptr_tbl[i]+ptr_size[i]) != END_CHAR ) {
			fprintf(stderr,"CORRUPT MEMORY: i=%5d  %s:%d nu=%5d size=%5d   addr=%lx - %lx\n",
				i,ptr_file[i],ptr_linenu[i],
				ptr_nu[i],ptr_size[i],ptr_tbl[i],
				ptr_tbl[i] +ptr_size[i] );
			fprintf(stderr,"CORRUPT MEMORY: last added = %d\n",alloc_nu-1);
			error = -1;
		}
	}
	if ( error ) {
		if ( make_core ) {
			core_program();
		}
		announce_bug("new_check - CORRUPT MEMORY\n");
	}
}



/*
#ifdef DOS
char *
bsearch(key, base, nelem, width, fcmp)
void *key;
void *base;
size_t nelem;
size_t width;
int (*fcmp)();
{
	size_t low, high, mid;
	int rc;

	low = 0;
	high = nelem;
	while ( low <= high ) {
		mid = (low+high)/2;
		rc = fcmp( (char *)key, (char *)base+mid*width);
		if ( rc == 0 ) return( (char *)base+mid*width );
		if ( rc < 0 ) high = mid - 1;
		else          low  = mid + 1;
	}
	return (NULL);
}
#endif
*/



char *
new_tr( file, linenu, size)
char *file;
int linenu;
unsigned size;
{
	char *A;
	filename = file;
	line_number = linenu;
	A = new(size);
	filename = "unknown";
	line_number = 0;

	return (A);
}

char *
rnew_tr( file, linenu, ptr, size)
char *file;
int linenu;
char *ptr;
unsigned size;
{
	char *A;

	filename = file;
	line_number = linenu;
	A = rnew(ptr,size);
	filename = "unknown";
	line_number = 0;
	return (A);
}

char *cnew_tr(file, linenu, n,size) 
char *file;
int linenu;
unsigned n,size;
{
	char *A;

	filename = file;
	line_number = linenu;
	A = cnew(n,size);
	filename = "unknown";
	line_number = 0;
	return (A);
}

void
sFree_tr(file, linenu, ptr) 
char *file;
int linenu;
char *ptr;
{
	filename = file;
	line_number = linenu;
	sFree(ptr);
	filename = "unknown";
	line_number = 0;
}

void
stop_if_nomem_tr(file, linenu, value)
char *file;
int linenu;
int value;
{
	filename = file;
	line_number = linenu;
	stop_if_nomem(value);
	filename = "unknown";
	line_number = 0;
}

static
void memory_nu(alloc_nu)
int alloc_nu;
{
	if ( alloc_nu == new_alloc_nu ) {
		if ( trace_func != NULL ) {
			(*trace_func)();
		}
	}

	if ( alloc_nu < -1 ) {
		fprintf(stderr,"Internal Error - alloc_nu = 5d\n",alloc_nu);
	}
}


void
new_print_memory_alloc()
{
	printf("Memory Allocated: %d\n",total);
}

void
print_all_new()
{
	if (trace>=1) print_new(-1);
}

static void
new_error(msg)
/*************************************************************/
/* New routine hit an error - print message then quit	     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
char *msg;
{
	fprintf(stderr,msg);
	if ( make_core ) {
		core_program();
	}
	exit_program(1);
}

